@@ -3,7 +3,7 @@
    DHCP Server Daemon. */
 
 /*
- * Copyright (c) 2004-2011 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2016 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -24,16 +24,10 @@
  *   <info@isc.org>
  *   https://www.isc.org/
  *
- * This software has been written for Internet Systems Consortium
- * by Ted Lemon in cooperation with Vixie Enterprises and Nominum, Inc.
- * To learn more about Internet Systems Consortium, see
- * ``https://www.isc.org/''.  To learn more about Vixie Enterprises,
- * see ``http://www.vix.com''.   To learn more about Nominum, Inc., see
- * ``http://www.nominum.com''.
  */
 
 static const char copyright[] =
-"Copyright 2004-2011 Internet Systems Consortium.";
+"Copyright 2004-2016 Internet Systems Consortium.";
 static const char arr [] = "All rights reserved.";
 static const char message [] = "Internet Systems Consortium DHCP Server";
 static const char url [] =
@@ -42,11 +36,11 @@ static const char url [] =
 #include "dhcpd.h"
 #include <omapip/omapip_p.h>
 #include <syslog.h>
+#include <signal.h>
 #include <errno.h>
 #include <limits.h>
 #include <sys/types.h>
 #include <sys/time.h>
-#include <signal.h>
 
 #if defined (PARANOIA)
 #  include <sys/types.h>
@@ -56,9 +50,15 @@ static const char url [] =
 #  define group real_group 
 #    include <grp.h>
 #  undef group
+
+/* global values so db.c can look at them */
+uid_t set_uid = 0;
+gid_t set_gid = 0;
 #endif /* PARANOIA */
 
+#ifndef UNIT_TEST
 static void usage(void);
+#endif
 
 struct iaddr server_identifier;
 int server_identifier_matched;
@@ -67,95 +67,17 @@ int server_identifier_matched;
 
 /* This stuff is always executed to figure the default values for certain
    ddns variables. */
-
 char std_nsupdate [] = "						    \n\
 option server.ddns-hostname =						    \n\
-  pick (option fqdn.hostname, option host-name);			    \n\
+  pick (option fqdn.hostname, option host-name, config-option host-name);   \n\
 option server.ddns-domainname =	config-option domain-name;		    \n\
 option server.ddns-rev-domainname = \"in-addr.arpa.\";";
 
-/* This is the old-style name service updater that is executed
-   whenever a lease is committed.  It does not follow the DHCP-DNS
-   draft at all. */
-
-char old_nsupdate [] = "						    \n\
-on commit {								    \n\
-  if (not static and							    \n\
-      ((config-option server.ddns-updates = null) or			    \n\
-       (config-option server.ddns-updates != 0))) {			    \n\
-    set new-ddns-fwd-name =						    \n\
-      concat (pick (config-option server.ddns-hostname,			    \n\
-		    option host-name), \".\",				    \n\
-	      pick (config-option server.ddns-domainname,		    \n\
-		    config-option domain-name));			    \n\
-    if (defined (ddns-fwd-name) and ddns-fwd-name != new-ddns-fwd-name) {   \n\
-      switch (ns-update (delete (IN, A, ddns-fwd-name, leased-address))) {  \n\
-      case NOERROR:							    \n\
-	unset ddns-fwd-name;						    \n\
-	on expiry or release {						    \n\
-	}								    \n\
-      }									    \n\
-    }									    \n\
-									    \n\
-    if (not defined (ddns-fwd-name)) {					    \n\
-      set ddns-fwd-name = new-ddns-fwd-name;				    \n\
-      if defined (ddns-fwd-name) {					    \n\
-	switch (ns-update (not exists (IN, A, ddns-fwd-name, null),	    \n\
-			   add (IN, A, ddns-fwd-name, leased-address,	    \n\
-				lease-time / 2))) {			    \n\
-	default:							    \n\
-	  unset ddns-fwd-name;						    \n\
-	  break;							    \n\
-									    \n\
-	case NOERROR:							    \n\
-	  set ddns-rev-name =						    \n\
-	    concat (binary-to-ascii (10, 8, \".\",			    \n\
-				     reverse (1,			    \n\
-					      leased-address)), \".\",	    \n\
-		    pick (config-option server.ddns-rev-domainname,	    \n\
-			  \"in-addr.arpa.\"));				    \n\
-	  switch (ns-update (delete (IN, PTR, ddns-rev-name, null),	    \n\
-			     add (IN, PTR, ddns-rev-name, ddns-fwd-name,    \n\
-				  lease-time / 2)))			    \n\
-	    {								    \n\
-	    default:							    \n\
-	      unset ddns-rev-name;					    \n\
-	      on release or expiry {					    \n\
-		switch (ns-update (delete (IN, A, ddns-fwd-name,	    \n\
-					   leased-address))) {		    \n\
-		case NOERROR:						    \n\
-		  unset ddns-fwd-name;					    \n\
-		  break;						    \n\
-		}							    \n\
-		on release or expiry;					    \n\
-	      }								    \n\
-	      break;							    \n\
-									    \n\
-	    case NOERROR:						    \n\
-	      on release or expiry {					    \n\
-		switch (ns-update (delete (IN, PTR, ddns-rev-name, null))) {\n\
-		case NOERROR:						    \n\
-		  unset ddns-rev-name;					    \n\
-		  break;						    \n\
-		}							    \n\
-		switch (ns-update (delete (IN, A, ddns-fwd-name,	    \n\
-					   leased-address))) {		    \n\
-		case NOERROR:						    \n\
-		  unset ddns-fwd-name;					    \n\
-		  break;						    \n\
-		}							    \n\
-		on release or expiry;					    \n\
-	      }								    \n\
-	    }								    \n\
-	}								    \n\
-      }									    \n\
-    }									    \n\
-    unset new-ddns-fwd-name;						    \n\
-  }									    \n\
-}";
-
 #endif /* NSUPDATE */
 int ddns_update_style;
+int dont_use_fsync = 0; /* 0 = default, use fsync, 1 = don't use fsync */
+int server_id_check = 0; /* 0 = default, don't check server id, 1 = do check */
+int prefix_length_mode = PLM_EXACT;
 
 const char *path_dhcpd_conf = _PATH_DHCPD_CONF;
 const char *path_dhcpd_db = _PATH_DHCPD_DB;
@@ -208,6 +130,12 @@ static void omapi_listener_start (void *foo)
 	omapi_object_dereference (&listener, MDL);
 }
 
+#ifndef UNIT_TEST
+
+/* Note: If we add unit tests to test setup_chroot it will
+ * need to be moved to be outside the ifndef UNIT_TEST block.
+ */
+
 #if defined (PARANOIA)
 /* to be used in one of two possible scenarios */
 static void setup_chroot (char *chroot_dir) {
@@ -224,7 +152,6 @@ static void setup_chroot (char *chroot_dir) {
 }
 #endif /* PARANOIA */
 
-#ifndef UNIT_TEST
 int 
 main(int argc, char **argv) {
 	int fd;
@@ -233,9 +160,9 @@ main(int argc, char **argv) {
 	char *s;
 	int cftest = 0;
 	int lftest = 0;
-#ifndef DEBUG
 	int pid;
 	char pbuf [20];
+#ifndef DEBUG
 	int daemon = 1;
 #endif
 	int quiet = 0;
@@ -262,9 +189,6 @@ main(int argc, char **argv) {
 	char *set_user   = 0;
 	char *set_group  = 0;
 	char *set_chroot = 0;
-
-	uid_t set_uid = 0;
-	gid_t set_gid = 0;
 #endif /* PARANOIA */
 
         /* Make sure that file descriptors 0 (stdin), 1, (stdout), and
@@ -281,7 +205,8 @@ main(int argc, char **argv) {
                 close(fd);
 
 	/* Set up the isc and dns library managers */
-	status = dhcp_context_create();
+	status = dhcp_context_create(DHCP_CONTEXT_PRE_DB,
+				     NULL, NULL);
 	if (status != ISC_R_SUCCESS)
 		log_fatal("Can't initialize context: %s",
 			  isc_result_totext(status));
@@ -302,7 +227,7 @@ main(int argc, char **argv) {
 	dhcp_common_objects_setup ();
 
 	/* Initially, log errors to stderr as well as to syslogd. */
-	openlog ("dhcpd", LOG_NDELAY, DHCPD_LOG_FACILITY);
+	openlog ("dhcpd", DHCP_LOG_OPTIONS, DHCPD_LOG_FACILITY);
 
 	for (i = 1; i < argc; i++) {
 		if (!strcmp (argv [i], "-p")) {
@@ -390,7 +315,13 @@ main(int argc, char **argv) {
 			local_family_set = 1;
 #endif /* DHCPv6 */
 		} else if (!strcmp (argv [i], "--version")) {
-			log_info("isc-dhcpd-%s", PACKAGE_VERSION);
+			const char vstring[] = "isc-dhcpd-";
+			IGNORE_RET(write(STDERR_FILENO, vstring,
+					 strlen(vstring)));
+			IGNORE_RET(write(STDERR_FILENO,
+					 PACKAGE_VERSION,
+					 strlen(PACKAGE_VERSION)));
+			IGNORE_RET(write(STDERR_FILENO, "\n", 1));
 			exit (0);
 #if defined (TRACING)
 		} else if (!strcmp (argv [i], "-tf")) {
@@ -464,12 +395,11 @@ main(int argc, char **argv) {
          * to be reopened after chdir() has been called
          */
         if (path_dhcpd_db[0] != '/') {
-                char *path = dmalloc(PATH_MAX, MDL);
-                if (path == NULL)
-                        log_fatal("No memory for filename\n");
-                path_dhcpd_db = realpath(path_dhcpd_db,  path);
+		const char *path = path_dhcpd_db;
+                path_dhcpd_db = realpath(path_dhcpd_db, NULL);
                 if (path_dhcpd_db == NULL)
-                        log_fatal("%s: %s", path, strerror(errno));
+                        log_fatal("Failed to get realpath for %s: %s", path, 
+                                   strerror(errno));
         }
 
 	if (!quiet) {
@@ -478,7 +408,6 @@ main(int argc, char **argv) {
 		log_info (arr);
 		log_info (url);
 	} else {
-		quiet = 0;
 		log_perror = 0;
 	}
 
@@ -495,7 +424,9 @@ main(int argc, char **argv) {
 	trace_srandom = trace_type_register ("random-seed", (void *)0,
 					     trace_seed_input,
 					     trace_seed_stop, MDL);
+#if defined (NSUPDATE)
 	trace_ddns_init();
+#endif /* NSUPDATE */
 #endif
 
 #if defined (PARANOIA)
@@ -624,6 +555,7 @@ main(int argc, char **argv) {
 	dhcp_interface_setup_hook = dhcpd_interface_setup_hook;
 	bootp_packet_handler = do_packet;
 #ifdef DHCPv6
+	add_enumeration (&prefix_length_modes);
 	dhcpv6_packet_handler = do_packet6;
 #endif /* DHCPv6 */
 
@@ -688,15 +620,48 @@ main(int argc, char **argv) {
 		log_fatal ("Configuration file errors encountered -- exiting");
 
 	postconf_initialization (quiet);
- 
+
 #if defined (PARANOIA) && !defined (EARLY_CHROOT)
 	if (set_chroot) setup_chroot (set_chroot);
 #endif /* PARANOIA && !EARLY_CHROOT */
 
+#ifdef DHCPv6
+	/* log info about ipv6_ponds with large address ranges */
+	report_jumbo_ranges();
+#endif
+
         /* test option should cause an early exit */
  	if (cftest && !lftest) 
  		exit(0);
 
+	/*
+	 * First part of dealing with pid files.  Check to see if
+	 * we should continue running or not.  We run if:
+	 * - we are testing the lease file out
+	 * - we don't have a pid file to check
+	 * - there is no other process running
+	 */
+	if ((lftest == 0) && (no_pid_file == ISC_FALSE)) {
+		/*Read previous pid file. */
+		if ((i = open(path_dhcpd_pid, O_RDONLY)) >= 0) {
+			status = read(i, pbuf, (sizeof pbuf) - 1);
+			close(i);
+			if (status > 0) {
+				pbuf[status] = 0;
+				pid = atoi(pbuf);
+
+				/*
+				 * If there was a previous server process and
+				 * it is still running, abort
+				 */
+				if (!pid ||
+				    (pid != getpid() && kill(pid, 0) == 0))
+					log_fatal("There's already a "
+						  "DHCP server running.");
+			}
+		}
+	}
+
 	group_write_hook = group_writer;
 
 	/* Start up the database... */
@@ -725,7 +690,6 @@ main(int argc, char **argv) {
 	}
 #endif /* DHCPv6 */
 
-
 	/* Make up a seed for the random number generator from current
 	   time plus the sum of the last four bytes of each
 	   interface's hardware address interpreted as an integer.
@@ -747,19 +711,21 @@ main(int argc, char **argv) {
 
 #ifdef DHCPv6
 	/*
-	 * Set server DHCPv6 identifier.
+	 * Set server DHCPv6 identifier - we go in order:
+	 * dhcp6.server-id in the config file
+	 * server-duid from the lease file
+	 * server-duid from the config file (the config file is read first
+	 * and the lease file overwrites the config file information)
+	 * genrate a new one
+	 * In all cases we write it out to the lease file.
 	 * See dhcpv6.c for discussion of setting DUID.
 	 */
-	if (set_server_duid_from_option() == ISC_R_SUCCESS) {
-		write_server_duid();
-	} else {
-		if (!server_duid_isset()) {
-			if (generate_new_server_duid() != ISC_R_SUCCESS) {
-				log_fatal("Unable to set server identifier.");
-			}
-			write_server_duid();
-		}
+	if ((set_server_duid_from_option() != ISC_R_SUCCESS) &&
+	    (!server_duid_isset()) &&
+	    (generate_new_server_duid() != ISC_R_SUCCESS)) {
+		log_fatal("Unable to set server identifier.");
 	}
+	write_server_duid();
 #endif /* DHCPv6 */
 
 #ifndef DEBUG
@@ -771,6 +737,23 @@ main(int argc, char **argv) {
 			exit (0);
 	}
  
+	/*
+	 * Second part of dealing with pid files.  Now
+	 * that we have forked we can write our pid if
+	 * appropriate.
+	 */
+	if (no_pid_file == ISC_FALSE) {
+		i = open(path_dhcpd_pid, O_WRONLY|O_CREAT|O_TRUNC, 0644);
+		if (i >= 0) {
+			sprintf(pbuf, "%d\n", (int) getpid());
+			IGNORE_RET(write(i, pbuf, strlen(pbuf)));
+			close(i);
+		} else {
+			log_error("Can't create PID file %s: %m.",
+				  path_dhcpd_pid);
+		}
+	}
+
 #if defined (PARANOIA)
 	/* change uid to the specified one */
 
@@ -787,42 +770,6 @@ main(int argc, char **argv) {
 	}
 #endif /* PARANOIA */
 
-	/*
-	 * Deal with pid files.  If the user told us
-	 * not to write a file we don't read one either
-	 */
-	if (no_pid_file == ISC_FALSE) {
-		/*Read previous pid file. */
-		if ((i = open (path_dhcpd_pid, O_RDONLY)) >= 0) {
-			status = read(i, pbuf, (sizeof pbuf) - 1);
-			close (i);
-			if (status > 0) {
-				pbuf[status] = 0;
-				pid = atoi(pbuf);
-
-				/*
-				 * If there was a previous server process and
-				 * it is still running, abort
-				 */
-				if (!pid ||
-				    (pid != getpid() && kill(pid, 0) == 0))
-					log_fatal("There's already a "
-						  "DHCP server running.");
-			}
-		}
-
-		/* Write new pid file. */
-		i = open(path_dhcpd_pid, O_WRONLY|O_CREAT|O_TRUNC, 0644);
-		if (i >= 0) {
-			sprintf(pbuf, "%d\n", (int) getpid());
-			IGNORE_RET (write(i, pbuf, strlen(pbuf)));
-			close(i);
-		} else {
-			log_error("Can't create PID file %s: %m.",
-				  path_dhcpd_pid);
-		}
-	}
-
 	/* If we were requested to log to stdout on the command line,
 	   keep doing so; otherwise, stop. */
 	if (log_perror == -1)
@@ -832,17 +779,17 @@ main(int argc, char **argv) {
 
 	if (daemon) {
 		/* Become session leader and get pid... */
-		pid = setsid();
+		(void) setsid();
 
                 /* Close standard I/O descriptors. */
-                close(0);
-                close(1);
-                close(2);
+                (void) close(0);
+                (void) close(1);
+                (void) close(2);
 
                 /* Reopen them on /dev/null. */
-                open("/dev/null", O_RDWR);
-                open("/dev/null", O_RDWR);
-                open("/dev/null", O_RDWR);
+                (void) open("/dev/null", O_RDWR);
+                (void) open("/dev/null", O_RDWR);
+                (void) open("/dev/null", O_RDWR);
                 log_perror = 0; /* No sense logging to /dev/null. */
 
        		IGNORE_RET (chdir("/"));
@@ -859,65 +806,70 @@ main(int argc, char **argv) {
 	omapi_set_int_value ((omapi_object_t *)dhcp_control_object,
 			     (omapi_object_t *)0, "state", server_running);
 
-	/* Receive packets and dispatch them... */
+#if defined(ENABLE_GENTLE_SHUTDOWN)
+	/* no signal handlers until we deal with the side effects */
+        /* install signal handlers */
+	signal(SIGINT, dhcp_signal_handler);   /* control-c */
+	signal(SIGTERM, dhcp_signal_handler);  /* kill */
+#endif
+
+	/* Log that we are about to start working */
+	log_info("Server starting service.");
+
+	/*
+	 * Receive packets and dispatch them...
+	 * dispatch() will never return.
+	 */
 	dispatch ();
 
-	/* Not reached */
+	/* Let's return status code */
 	return 0;
 }
 #endif /* !UNIT_TEST */
 
 void postconf_initialization (int quiet)
 {
-	struct option_state *options = (struct option_state *)0;
+	struct option_state *options = NULL;
 	struct data_string db;
 	struct option_cache *oc;
 	char *s;
 	isc_result_t result;
-#if defined (NSUPDATE)
-	struct parse *parse;
-#endif
 	int tmp;
+#if defined (NSUPDATE)
+	struct in_addr  local4, *local4_ptr = NULL;
+	struct in6_addr local6, *local6_ptr = NULL;
+#endif
 
 	/* Now try to get the lease file name. */
-	option_state_allocate (&options, MDL);
+	option_state_allocate(&options, MDL);
 
-	execute_statements_in_scope ((struct binding_value **)0,
-				     (struct packet *)0,
-				     (struct lease *)0,
-				     (struct client_state *)0,
-				     (struct option_state *)0,
-				     options, &global_scope,
-				     root_group,
-				     (struct group *)0);
-	memset (&db, 0, sizeof db);
-	oc = lookup_option (&server_universe, options, SV_LEASE_FILE_NAME);
+	execute_statements_in_scope(NULL, NULL, NULL, NULL, NULL,
+				    options, &global_scope, root_group,
+				    NULL, NULL);
+	memset(&db, 0, sizeof db);
+	oc = lookup_option(&server_universe, options, SV_LEASE_FILE_NAME);
 	if (oc &&
-	    evaluate_option_cache (&db, (struct packet *)0,
-				   (struct lease *)0, (struct client_state *)0,
-				   options, (struct option_state *)0,
-				   &global_scope, oc, MDL)) {
-		s = dmalloc (db.len + 1, MDL);
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+				  &global_scope, oc, MDL)) {
+		s = dmalloc(db.len + 1, MDL);
 		if (!s)
-			log_fatal ("no memory for lease db filename.");
-		memcpy (s, db.data, db.len);
-		s [db.len] = 0;
-		data_string_forget (&db, MDL);
+			log_fatal("no memory for lease db filename.");
+		memcpy(s, db.data, db.len);
+		s[db.len] = 0;
+		data_string_forget(&db, MDL);
 		path_dhcpd_db = s;
 	}
 
-	oc = lookup_option (&server_universe, options, SV_PID_FILE_NAME);
+	oc = lookup_option(&server_universe, options, SV_PID_FILE_NAME);
 	if (oc &&
-	    evaluate_option_cache (&db, (struct packet *)0,
-				   (struct lease *)0, (struct client_state *)0,
-				   options, (struct option_state *)0,
-				   &global_scope, oc, MDL)) {
-		s = dmalloc (db.len + 1, MDL);
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+				  &global_scope, oc, MDL)) {
+		s = dmalloc(db.len + 1, MDL);
 		if (!s)
-			log_fatal ("no memory for pid filename.");
-		memcpy (s, db.data, db.len);
-		s [db.len] = 0;
-		data_string_forget (&db, MDL);
+			log_fatal("no memory for pid filename.");
+		memcpy(s, db.data, db.len);
+		s[db.len] = 0;
+		data_string_forget(&db, MDL);
 		path_dhcpd_pid = s;
 	}
 
@@ -930,137 +882,116 @@ void postconf_initialization (int quiet)
                 oc = lookup_option(&server_universe, options,
                                    SV_DHCPV6_LEASE_FILE_NAME);
                 if (oc &&
-                    evaluate_option_cache(&db, NULL, NULL, NULL,
-				          options, NULL, &global_scope,
-                                          oc, MDL)) {
-                        s = dmalloc (db.len + 1, MDL);
+                    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+					  &global_scope, oc, MDL)) {
+                        s = dmalloc(db.len + 1, MDL);
                         if (!s)
-                                log_fatal ("no memory for lease db filename.");
-                        memcpy (s, db.data, db.len);
-                        s [db.len] = 0;
-                        data_string_forget (&db, MDL);
+                                log_fatal("no memory for lease db filename.");
+                        memcpy(s, db.data, db.len);
+                        s[db.len] = 0;
+                        data_string_forget(&db, MDL);
                         path_dhcpd_db = s;
                 }
 
                 oc = lookup_option(&server_universe, options,
                                    SV_DHCPV6_PID_FILE_NAME);
                 if (oc &&
-                    evaluate_option_cache(&db, NULL, NULL, NULL,
-				          options, NULL, &global_scope,
-                                          oc, MDL)) {
-                        s = dmalloc (db.len + 1, MDL);
+                    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+					  &global_scope, oc, MDL)) {
+                        s = dmalloc(db.len + 1, MDL);
                         if (!s)
-                                log_fatal ("no memory for pid filename.");
-                        memcpy (s, db.data, db.len);
-                        s [db.len] = 0;
-                        data_string_forget (&db, MDL);
+                                log_fatal("no memory for pid filename.");
+                        memcpy(s, db.data, db.len);
+                        s[db.len] = 0;
+                        data_string_forget(&db, MDL);
                         path_dhcpd_pid = s;
                 }
         }
 #endif /* DHCPv6 */
 
 	omapi_port = -1;
-	oc = lookup_option (&server_universe, options, SV_OMAPI_PORT);
+	oc = lookup_option(&server_universe, options, SV_OMAPI_PORT);
 	if (oc &&
-	    evaluate_option_cache (&db, (struct packet *)0,
-				   (struct lease *)0, (struct client_state *)0,
-				   options, (struct option_state *)0,
-				   &global_scope, oc, MDL)) {
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+				  &global_scope, oc, MDL)) {
 		if (db.len == 2) {
-			omapi_port = getUShort (db.data);
+			omapi_port = getUShort(db.data);
 		} else
-			log_fatal ("invalid omapi port data length");
-		data_string_forget (&db, MDL);
+			log_fatal("invalid omapi port data length");
+		data_string_forget(&db, MDL);
 	}
 
-	oc = lookup_option (&server_universe, options, SV_OMAPI_KEY);
+	oc = lookup_option(&server_universe, options, SV_OMAPI_KEY);
 	if (oc &&
-	    evaluate_option_cache (&db, (struct packet *)0,
-				   (struct lease *)0, (struct client_state *)0,
-				   options,
-				   (struct option_state *)0,
-				   &global_scope, oc, MDL)) {
-		s = dmalloc (db.len + 1, MDL);
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+				  &global_scope, oc, MDL)) {
+		s = dmalloc(db.len + 1, MDL);
 		if (!s)
-			log_fatal ("no memory for OMAPI key filename.");
-		memcpy (s, db.data, db.len);
-		s [db.len] = 0;
-		data_string_forget (&db, MDL);
-		result = omapi_auth_key_lookup_name (&omapi_key, s);
-		dfree (s, MDL);
+			log_fatal("no memory for OMAPI key filename.");
+		memcpy(s, db.data, db.len);
+		s[db.len] = 0;
+		data_string_forget(&db, MDL);
+		result = omapi_auth_key_lookup_name(&omapi_key, s);
+		dfree(s, MDL);
 		if (result != ISC_R_SUCCESS)
-			log_fatal ("OMAPI key %s: %s",
-				   s, isc_result_totext (result));
+			log_fatal("OMAPI key %s: %s",
+				  s, isc_result_totext (result));
 	}
 
-	oc = lookup_option (&server_universe, options, SV_LOCAL_PORT);
+	oc = lookup_option(&server_universe, options, SV_LOCAL_PORT);
 	if (oc &&
-	    evaluate_option_cache (&db, (struct packet *)0,
-				   (struct lease *)0, (struct client_state *)0,
-				   options,
-				   (struct option_state *)0,
-				   &global_scope, oc, MDL)) {
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+				  &global_scope, oc, MDL)) {
 		if (db.len == 2) {
-			local_port = htons (getUShort (db.data));
+			local_port = htons(getUShort (db.data));
 		} else
-			log_fatal ("invalid local port data length");
-		data_string_forget (&db, MDL);
+			log_fatal("invalid local port data length");
+		data_string_forget(&db, MDL);
 	}
 
-	oc = lookup_option (&server_universe, options, SV_REMOTE_PORT);
+	oc = lookup_option(&server_universe, options, SV_REMOTE_PORT);
 	if (oc &&
-	    evaluate_option_cache (&db, (struct packet *)0,
-				   (struct lease *)0, (struct client_state *)0,
-				   options, (struct option_state *)0,
-				   &global_scope, oc, MDL)) {
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+				  &global_scope, oc, MDL)) {
 		if (db.len == 2) {
-			remote_port = htons (getUShort (db.data));
+			remote_port = htons(getUShort (db.data));
 		} else
-			log_fatal ("invalid remote port data length");
-		data_string_forget (&db, MDL);
+			log_fatal("invalid remote port data length");
+		data_string_forget(&db, MDL);
 	}
 
-	oc = lookup_option (&server_universe, options,
-			    SV_LIMITED_BROADCAST_ADDRESS);
+	oc = lookup_option(&server_universe, options,
+			   SV_LIMITED_BROADCAST_ADDRESS);
 	if (oc &&
-	    evaluate_option_cache (&db, (struct packet *)0,
-				   (struct lease *)0, (struct client_state *)0,
-				   options, (struct option_state *)0,
-				   &global_scope, oc, MDL)) {
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+				  &global_scope, oc, MDL)) {
 		if (db.len == 4) {
-			memcpy (&limited_broadcast, db.data, 4);
+			memcpy(&limited_broadcast, db.data, 4);
 		} else
-			log_fatal ("invalid broadcast address data length");
-		data_string_forget (&db, MDL);
+			log_fatal("invalid broadcast address data length");
+		data_string_forget(&db, MDL);
 	}
 
-	oc = lookup_option (&server_universe, options,
-			    SV_LOCAL_ADDRESS);
+	oc = lookup_option(&server_universe, options, SV_LOCAL_ADDRESS);
 	if (oc &&
-	    evaluate_option_cache (&db, (struct packet *)0,
-				   (struct lease *)0, (struct client_state *)0,
-				   options, (struct option_state *)0,
-				   &global_scope, oc, MDL)) {
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+				  &global_scope, oc, MDL)) {
 		if (db.len == 4) {
-			memcpy (&local_address, db.data, 4);
+			memcpy(&local_address, db.data, 4);
 		} else
-			log_fatal ("invalid local address data length");
-		data_string_forget (&db, MDL);
+			log_fatal("invalid local address data length");
+		data_string_forget(&db, MDL);
 	}
 
-	oc = lookup_option (&server_universe, options, SV_DDNS_UPDATE_STYLE);
+	oc = lookup_option(&server_universe, options, SV_DDNS_UPDATE_STYLE);
 	if (oc) {
-		if (evaluate_option_cache (&db, (struct packet *)0,
-					   (struct lease *)0,
-					   (struct client_state *)0,
-					   options,
-					   (struct option_state *)0,
-					   &global_scope, oc, MDL)) {
+		if (evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+					  &global_scope, oc, MDL)) {
 			if (db.len == 1) {
-				ddns_update_style = db.data [0];
+				ddns_update_style = db.data[0];
 			} else
-				log_fatal ("invalid dns update type");
-			data_string_forget (&db, MDL);
+				log_fatal("invalid dns update type");
+			data_string_forget(&db, MDL);
 		}
 	} else {
 		ddns_update_style = DDNS_UPDATE_STYLE_NONE;
@@ -1070,6 +1001,35 @@ void postconf_initialization (int quiet)
 	if (ddns_update_style == DDNS_UPDATE_STYLE_AD_HOC) {
 		log_fatal("ddns-update-style ad_hoc no longer supported");
 	}
+
+	oc = lookup_option(&server_universe, options, SV_DDNS_LOCAL_ADDRESS4);
+	if (oc) {
+		if (evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+					  &global_scope, oc, MDL)) {
+			if (db.len == 4) {
+				memcpy(&local4, db.data, 4);
+				local4_ptr = &local4;
+			}
+			data_string_forget(&db, MDL);
+		}
+	}
+
+	oc = lookup_option(&server_universe, options, SV_DDNS_LOCAL_ADDRESS6);
+	if (oc) {
+		if (evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+					  &global_scope, oc, MDL)) {
+			if (db.len == 16) {
+				memcpy(&local6, db.data, 16);
+				local6_ptr = &local6;
+			}
+			data_string_forget(&db, MDL);
+		}
+	}
+
+	if (dhcp_context_create(DHCP_CONTEXT_POST_DB, local4_ptr, local6_ptr)
+	    != ISC_R_SUCCESS)
+		log_fatal("Unable to complete ddns initialization");
+
 #else
 	/* If we don't have support for updates compiled in tell the user */
 	if (ddns_update_style != DDNS_UPDATE_STYLE_NONE) {
@@ -1077,36 +1037,36 @@ void postconf_initialization (int quiet)
 	}
 #endif
 
-	oc = lookup_option (&server_universe, options, SV_LOG_FACILITY);
+	if (!quiet) {
+		log_info ("Config file: %s", path_dhcpd_conf);
+		log_info ("Database file: %s", path_dhcpd_db);
+		log_info ("PID file: %s", path_dhcpd_pid);
+	}
+
+	oc = lookup_option(&server_universe, options, SV_LOG_FACILITY);
 	if (oc) {
-		if (evaluate_option_cache (&db, (struct packet *)0,
-					   (struct lease *)0,
-					   (struct client_state *)0,
-					   options,
-					   (struct option_state *)0,
-					   &global_scope, oc, MDL)) {
+		if (evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+					  &global_scope, oc, MDL)) {
 			if (db.len == 1) {
 				closelog ();
-				openlog ("dhcpd", LOG_NDELAY, db.data[0]);
+				openlog("dhcpd", DHCP_LOG_OPTIONS, db.data[0]);
 				/* Log the startup banner into the new
 				   log file. */
-				if (!quiet) {
-					/* Don't log to stderr twice. */
-					tmp = log_perror;
-					log_perror = 0;
-					log_info("%s %s",
-						 message, PACKAGE_VERSION);
-					log_info (copyright);
-					log_info (arr);
-					log_info (url);
-					log_perror = tmp;
-				}
+				/* Don't log to stderr twice. */
+				tmp = log_perror;
+				log_perror = 0;
+				log_info("%s %s", message, PACKAGE_VERSION);
+				log_info(copyright);
+				log_info(arr);
+				log_info(url);
+				log_perror = tmp;
 			} else
-				log_fatal ("invalid log facility");
-			data_string_forget (&db, MDL);
+				log_fatal("invalid log facility");
+			data_string_forget(&db, MDL);
 		}
 	}
-	
+
+#if defined(DELAYED_ACK)
 	oc = lookup_option(&server_universe, options, SV_DELAYED_ACK);
 	if (oc &&
 	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
@@ -1134,52 +1094,45 @@ void postconf_initialization (int quiet)
 
 		data_string_forget(&db, MDL);
 	}
+#endif
 
-	/* Don't need the options anymore. */
-	option_state_dereference (&options, MDL);
-	
-#if defined (NSUPDATE)
-	/* If old-style ddns updates have been requested, parse the
-	   old-style ddns updater. */
-	if (ddns_update_style == 1) {
-		struct executable_statement **e, *s;
+	oc = lookup_option(&server_universe, options, SV_DONT_USE_FSYNC);
+	if ((oc != NULL) &&
+	    evaluate_boolean_option_cache(NULL, NULL, NULL, NULL, options, NULL,
+					  &global_scope, oc, MDL)) {
+		dont_use_fsync = 1;
+		log_error("Not using fsync() to flush lease writes");
+	}
 
-		if (root_group -> statements) {
-			s = (struct executable_statement *)0;
-			if (!executable_statement_allocate (&s, MDL))
-				log_fatal ("no memory for ddns updater");
-			executable_statement_reference
-				(&s -> next, root_group -> statements, MDL);
-			executable_statement_dereference
-				(&root_group -> statements, MDL);
-			executable_statement_reference
-				(&root_group -> statements, s, MDL);
-			s -> op = statements_statement;
-			e = &s -> data.statements;
-			executable_statement_dereference (&s, MDL);
+       oc = lookup_option(&server_universe, options, SV_SERVER_ID_CHECK);
+       if ((oc != NULL) &&
+	   evaluate_boolean_option_cache(NULL, NULL, NULL, NULL, options, NULL,
+					 &global_scope, oc, MDL)) {
+		log_info("Setting server-id-check true");
+		server_id_check = 1;
+	}
+
+	oc = lookup_option(&server_universe, options, SV_PREFIX_LEN_MODE);
+	if ((oc != NULL) && 
+	    evaluate_option_cache(&db, NULL, NULL, NULL, options, NULL,
+					  &global_scope, oc, MDL)) {
+		if (db.len == 1) {
+			prefix_length_mode = db.data[0];
 		} else {
-			e = &root_group -> statements;
+			log_fatal("invalid prefix-len-mode");
 		}
 
-		/* Set up the standard name service updater routine. */
-		parse = NULL;
-		result = new_parse(&parse, -1, old_nsupdate,
-				   sizeof(old_nsupdate) - 1,
-				   "old name service update routine", 0);
-		if (result != ISC_R_SUCCESS)
-			log_fatal ("can't begin parsing old ddns updater!");
+		data_string_forget(&db, MDL);
+	}
 
-		if (parse != NULL) {
-			tmp = 0;
-			if (!(parse_executable_statements(e, parse, &tmp,
-							  context_any))) {
-				end_parse(&parse);
-				log_fatal("can't parse standard ddns updater!");
-			}
-		}
-		end_parse(&parse);
+#if defined (BINARY_LEASES)
+	if (local_family == AF_INET) {
+		log_info("Source compiled to use binary-leases");
 	}
 #endif
+
+	/* Don't need the options anymore. */
+	option_state_dereference(&options, MDL);
 }
 
 void postdb_startup (void)
@@ -1201,7 +1154,7 @@ void postdb_startup (void)
 }
 
 /* Print usage message. */
-
+#ifndef UNIT_TEST
 static void
 usage(void) {
 	log_info("%s %s", message, PACKAGE_VERSION);
@@ -1225,6 +1178,7 @@ usage(void) {
 		  "             [-pf pid-file] [--no-pid] [-s server]\n"
 		  "             [if0 [...ifN]]");
 }
+#endif
 
 void lease_pinged (from, packet, length)
 	struct iaddr from;
@@ -1463,6 +1417,8 @@ static isc_result_t dhcp_io_shutdown_countdown (void *vlp)
 	    free_everything ();
 	    omapi_print_dmalloc_usage_by_caller ();
 #endif
+	    if (no_pid_file == ISC_FALSE)
+		    (void) unlink(path_dhcpd_pid);
 	    exit (0);
 	}		
 #else
@@ -1472,6 +1428,8 @@ static isc_result_t dhcp_io_shutdown_countdown (void *vlp)
 		free_everything ();
 		omapi_print_dmalloc_usage_by_caller (); 
 #endif
+		if (no_pid_file == ISC_FALSE)
+			(void) unlink(path_dhcpd_pid);
 		exit (0);
 	}
 #endif
@@ -1494,11 +1452,32 @@ static isc_result_t dhcp_io_shutdown_countdown (void *vlp)
 isc_result_t dhcp_set_control_state (control_object_state_t oldstate,
 				     control_object_state_t newstate)
 {
-	if (newstate == server_shutdown) {
-		shutdown_time = cur_time;
-		shutdown_state = shutdown_listeners;
+	struct timeval tv;
+
+	if (newstate != server_shutdown)
+		return DHCP_R_INVALIDARG;
+	/* Re-entry. */
+	if (shutdown_signal == SIGUSR1)
+		return ISC_R_SUCCESS;
+	shutdown_time = cur_time;
+	shutdown_state = shutdown_listeners;
+	/* Called by user. */
+	if (shutdown_signal == 0) {
+		shutdown_signal = SIGUSR1;
 		dhcp_io_shutdown_countdown (0);
 		return ISC_R_SUCCESS;
 	}
-	return DHCP_R_INVALIDARG;
+	/* Called on signal. */
+	log_info("Received signal %d, initiating shutdown.", shutdown_signal);
+	shutdown_signal = SIGUSR1;
+	
+	/*
+	 * Prompt the shutdown event onto the timer queue
+	 * and return to the dispatch loop.
+	 */
+	tv.tv_sec = cur_tv.tv_sec;
+	tv.tv_usec = cur_tv.tv_usec + 1;
+	add_timeout(&tv,
+		    (void (*)(void *))dhcp_io_shutdown_countdown, 0, 0, 0);
+	return ISC_R_SUCCESS;
 }
